use std::collections::BTreeMap;

use hercules_ir::*;

use crate::*;

/*
 * The SLF lattice tracks what sub-values of a collection are known. Each sub-
 * value is a node ID at a set of indices that were written at. A write to a set
 * of indices that structurally maps a previous sub-value removes the old sub-
 * value, since that write may overwrite the old known sub-value. The lattice
 * top corresponds to every value is 0. When the sub-values at a set of indices
 * are not known, the `subvalues` map stores `None` for the known value. When a
 * write involves array positions, remove sub-values that are clobbered and
 * insert an indices set with an empty positions list and a `None` value.
 */
#[derive(Debug, Clone, PartialEq, Eq, Hash)]
struct SLFLattice {
    subvalues: BTreeMap<Box<[Index]>, Option<NodeID>>,
}

impl Semilattice for SLFLattice {
    fn meet(a: &Self, b: &Self) -> Self {
        // Merge the two maps. Find equal indices sets between `a` and `b` and
        // keep their known sub-value if they're equivalent. All other indices
        // sets in `a` or `b` map to `None`.
        let mut ret = BTreeMap::new();
        for (indices, a_subvalue) in a.subvalues.iter() {
            if let Some(b_subvalue) = b.subvalues.get(indices)
                && a_subvalue == b_subvalue
            {
                // If both maps have the same sub-value for this set of indices,
                // add it unmolested to the meet lattice value.
                ret.insert(indices.clone(), *a_subvalue);
            } else {
                // If not both maps have a write at the same set of indices or
                // if the writes don't match, then we don't know what's been
                // written there.
                ret.insert(indices.clone(), None);
            }
        }
        for (indices, _) in b.subvalues.iter() {
            // Any indices sets in `b` that aren't in `ret` are indices sets
            // that aren't in `a`, so the sub-value isn't known.
            ret.entry(indices.clone()).or_insert(None);
        }
        SLFLattice { subvalues: ret }
    }

    fn top() -> Self {
        SLFLattice {
            subvalues: BTreeMap::new(),
        }
    }

    fn bottom() -> Self {
        let mut subvalues = BTreeMap::new();
        // The empty indices set overlaps with all possible indices sets.
        subvalues.insert(Box::new([]) as Box<[Index]>, None);
        SLFLattice { subvalues }
    }
}

/*
 * Top level function to run store-to-load forwarding on a function. Looks for
 * known values inside collections and replaces reads of those values with the
 * values directly.
 */
pub fn slf(editor: &mut FunctionEditor, reverse_postorder: &Vec<NodeID>, typing: &Vec<TypeID>) {
    // First, run a dataflow analysis that looks at known values inside
    // collections. Thanks to the value semantics of Hercules IR, this analysis
    // is relatively simple and straightforward.
    let func = editor.func();
    let lattice = forward_dataflow(func, reverse_postorder, |inputs, id| {
        match func.nodes[id.idx()] {
            Node::Phi {
                control: _,
                data: _,
            }
            | Node::Reduce {
                control: _,
                init: _,
                reduct: _,
            }
            | Node::Ternary {
                op: TernaryOperator::Select,
                first: _,
                second: _,
                third: _,
            } => inputs.into_iter().fold(SLFLattice::top(), |acc, input| {
                SLFLattice::meet(&acc, input)
            }),
            Node::Write {
                collect: _,
                data,
                ref indices,
            } => {
                // Start with the indices of the `collect` input.
                let mut value = inputs[0].clone();

                // Any indices sets that overlap with `indices` become `None`,
                // since we no longer know what's stored there.
                for (other_indices, subvalue) in value.subvalues.iter_mut() {
                    if indices_may_overlap(other_indices, indices) {
                        *subvalue = None;
                    }
                }

                // Track `data` at `indices`.
                value.subvalues.insert(indices.clone(), Some(data));

                value
            }
            _ => SLFLattice::bottom(),
        }
    });

    // Second, look for reads where the indices set either:
    // 1. Equal the indices of a known sub-value. Then, the read can be replaced
    //    by the known sub-value.
    // 2. Otherwise, if the indices set doesn't overlap with any known or
    //    unknown sub-value, then the read can be replaced by a zero constant.
    // 3. Otherwise, the read can't be replaced.
    // Keep track of which nodes we've already replaced, since a sub-value we
    // knew previously may be the ID of an old node replaced previously.
    let mut replacements = BTreeMap::new();
    for id in editor.node_ids() {
        let Node::Read {
            collect,
            ref indices,
        } = editor.func().nodes[id.idx()]
        else {
            continue;
        };
        let subvalues = &lattice[collect.idx()].subvalues;

        if let Some(sub_value) = subvalues.get(indices)
            && let Some(mut known) = *sub_value
        {
            while let Some(replacement) = replacements.get(&known) {
                known = *replacement;
            }
            editor.edit(|mut edit| {
                edit = edit.replace_all_uses(id, known)?;
                edit.delete_node(id)
            });
            replacements.insert(id, known);
        } else if !subvalues
            .keys()
            .any(|other_indices| indices_may_overlap(other_indices, indices))
        {
            editor.edit(|mut edit| {
                let zero = edit.add_zero_constant(typing[id.idx()]);
                let zero = edit.add_node(Node::Constant { id: zero });
                edit = edit.replace_all_uses(id, zero)?;
                edit.delete_node(id)
            });
        }
    }
}
